# SPDX-License-Identifier: BSD-2

from .internal.utils import _chkrc, _lib_version_atleast
from ._libtpm2_pytss import ffi, lib
from .constants import TSS2_RC, TPM2_RC
from .TSS2_Exception import TSS2_Exception
from .TCTI import TCTI

if not _lib_version_atleast("tss2-tcti-spi-helper", "0.0.0"):
    raise NotImplementedError("Package tss2-tcti-spi-helper not present")


@ffi.def_extern()
def _tcti_spi_helper_sleep_ms(userdata, milliseconds):
    thiz = TCTISPIHelper._cffi_cast(userdata)
    if not hasattr(thiz, "on_sleep_ms"):
        return TSS2_RC.TCTI_RC_NOT_IMPLEMENTED
    try:
        thiz.on_sleep_ms(milliseconds)
    except Exception as e:
        rc = e.rc if isinstance(e, TSS2_Exception) else TSS2_RC.TCTI_RC_GENERAL_FAILURE
        thiz._set_last_exception(e)
        return rc

    return TPM2_RC.SUCCESS


@ffi.def_extern()
def _tcti_spi_helper_start_timeout(userdata, milliseconds):
    thiz = TCTISPIHelper._cffi_cast(userdata)
    if not hasattr(thiz, "on_start_timeout"):
        return TSS2_RC.TCTI_RC_NOT_IMPLEMENTED
    try:
        thiz.on_start_timeout(milliseconds)
    except Exception as e:
        rc = e.rc if isinstance(e, TSS2_Exception) else TSS2_RC.TCTI_RC_GENERAL_FAILURE
        thiz._set_last_exception(e)
        return rc

    return TPM2_RC.SUCCESS


@ffi.def_extern()
def _tcti_spi_helper_timeout_expired(userdata, is_time_expired) -> bool:

    thiz = TCTISPIHelper._cffi_cast(userdata)
    if not hasattr(thiz, "on_timeout_expired"):
        return TSS2_RC.TCTI_RC_NOT_IMPLEMENTED
    try:
        result = thiz.on_timeout_expired()
        is_time_expired[0] = result
    except Exception as e:
        rc = e.rc if isinstance(e, TSS2_Exception) else TSS2_RC.TCTI_RC_GENERAL_FAILURE
        thiz._set_last_exception(e)
        return rc

    return TPM2_RC.SUCCESS


@ffi.def_extern()
def _tcti_spi_helper_spi_acquire(userdata):
    thiz = TCTISPIHelper._cffi_cast(userdata)
    if not hasattr(thiz, "on_start_timeout"):
        return TSS2_RC.TCTI_RC_NOT_IMPLEMENTED
    try:
        thiz.on_spi_acquire()
    except Exception as e:
        rc = e.rc if isinstance(e, TSS2_Exception) else TSS2_RC.TCTI_RC_GENERAL_FAILURE
        thiz._set_last_exception(e)
        return rc

    return TPM2_RC.SUCCESS


@ffi.def_extern()
def _tcti_spi_helper_spi_release(userdata):
    thiz = TCTISPIHelper._cffi_cast(userdata)
    if not hasattr(thiz, "on_spi_release"):
        return TSS2_RC.TCTI_RC_NOT_IMPLEMENTED
    try:
        thiz.on_spi_release()
    except Exception as e:
        rc = e.rc if isinstance(e, TSS2_Exception) else TSS2_RC.TCTI_RC_GENERAL_FAILURE
        thiz._set_last_exception(e)
        return rc

    return TPM2_RC.SUCCESS


@ffi.def_extern()
def _tcti_spi_helper_spi_transfer(userdata, data_out, data_in, cnt):
    thiz = TCTISPIHelper._cffi_cast(userdata)
    if not hasattr(thiz, "on_spi_transfer"):
        return TSS2_RC.TCTI_RC_NOT_IMPLEMENTED
    try:
        # setup for the transaction
        dout = None if data_out == ffi.NULL else bytes(ffi.buffer(data_out, cnt))

        # call for transaction
        data_got = thiz.on_spi_transfer(dout)

        # handle response should None be OK?
        if data_got is None and data_in != ffi.NULL:
            raise RuntimeError("Response data CANNOT be None")
        elif data_got is None and data_in == ffi.NULL:
            return TPM2_RC.SUCCESS

        # current interface is hardcoded to full duplex, so input
        # must equal output
        if len(data_got) != cnt:
            raise ValueError(
                f"Transactions is expected to be {cnt} bytes, got {len(data_got)} bytes"
            )

        # copy the data
        raw_data_got = ffi.from_buffer(data_got)
        ffi.memmove(data_in, raw_data_got, len(data_got))

    except Exception as e:
        rc = e.rc if isinstance(e, TSS2_Exception) else TSS2_RC.TCTI_RC_GENERAL_FAILURE
        thiz._set_last_exception(e)
        return rc

    return TPM2_RC.SUCCESS


@ffi.def_extern()
def _tcti_spi_helper_finalize(userdata):
    thiz = TCTISPIHelper._cffi_cast(userdata)
    if hasattr(thiz, "on_finalize"):
        thiz.on_finalize(thiz)


class TCTISPIHelper(TCTI):
    """The  TCTI for interacting with SPI devices.

    Users should *extend* a TCTISPIHelper object and implement the following callbacks:

    All Users:
      - on_sleep_ms
      - on_start_timeout
      - on_timeout_expired
      - on_spi_transfer

    with_wait_state=true:
      - on_spi_acquire
      - on_spi_release

    Optional:
      - on_finalize

    Args:
        with_wait_state (bool): True if you intend to use wait states. Defaults to False.
    """

    def __init__(self, with_wait_state=False):
        self._with_wait_state = with_wait_state

        size = ffi.new("size_t *")
        self._callbacks = ffi.new("TSS2_TCTI_SPI_HELPER_PLATFORM *")
        self._callbacks.sleep_ms = lib._tcti_spi_helper_sleep_ms
        self._callbacks.start_timeout = lib._tcti_spi_helper_start_timeout
        self._callbacks.timeout_expired = lib._tcti_spi_helper_timeout_expired
        self._callbacks.spi_transfer = lib._tcti_spi_helper_spi_transfer
        self._callbacks.finalize = lib._tcti_spi_helper_finalize
        self._callbacks.user_data = self._thiz = ffi.new_handle(self)

        missing_implementation = []
        if self._with_wait_state:
            self._callbacks.spi_acquire = lib._tcti_spi_helper_spi_acquire
            self._callbacks.spi_release = lib._tcti_spi_helper_spi_release
            if "TCTISPIHelper.on_spi_acquire" in str(self.on_spi_acquire):
                missing_implementation.append("on_spi_acquire")

            if "TCTISPIHelper.on_spi_release" in str(self.on_spi_release):
                missing_implementation.append("on_spi_release")

        if "TCTISPIHelper.on_spi_transfer" in str(self.on_spi_transfer):
            missing_implementation.append("on_spi_transfer")

        if "TCTISPIHelper.on_timeout_expired" in str(self.on_timeout_expired):
            missing_implementation.append("on_timeout_expired")

        if "TCTISPIHelper.on_start_timeout" in str(self.on_start_timeout):
            missing_implementation.append("on_start_timeout")

        if "TCTISPIHelper.on_sleep_ms" in str(self.on_sleep_ms):
            missing_implementation.append("on_sleep_ms")

        if len(missing_implementation) > 0:
            raise NotImplementedError(
                f"Subclasses must implement {','.join(missing_implementation)}"
            )

        _chkrc(lib.Tss2_Tcti_Spi_Helper_Init(ffi.NULL, size, self._callbacks))

        self._tcti_mem = ffi.new(f"uint8_t [{size[0]}]")
        self._opaque_tcti_ctx = ffi.cast("TSS2_TCTI_CONTEXT *", self._tcti_mem)

        try:
            self._clear_exceptions()
            _chkrc(
                lib.Tss2_Tcti_Spi_Helper_Init(
                    self._opaque_tcti_ctx, size, self._callbacks
                )
            )
        except Exception as e:
            e = self._get_current_exception(e)
            self._clear_exceptions()
            raise e

        super().__init__(self._opaque_tcti_ctx)

    @property
    def waitstate(self):
        """Gets the wait state property.

        Returns(bool):
            True if this TCTI implements wait states, false otherwise.
        """
        return self._with_wait_state

    @staticmethod
    def _cffi_cast(userdata):
        return ffi.from_handle(userdata)

    def on_sleep_ms(self, milliseconds: int) -> None:
        """Sleeps for a specified amount of time in millisecons.

        This callback is REQUIRED.

        Args:
            milliseconds(int): The time to sleep.

        Raises:
            Exception: Implementations are free to raise any Exception. Exceptions are retained
            across the native boundary.
        """
        pass

    def on_start_timeout(self, milliseconds: int) -> None:
        """Called when a timeout is occurring with the sleep duration in millisecons.

        This callback is REQUIRED.

        Args:
            milliseconds(int): The time to sleep.

        Raises:
            Exception: Implementations are free to raise any Exception. Exceptions are retained
            across the native boundary.
        """
        pass

    def on_timeout_expired(self) -> bool:
        """Called to determine if a timeout is expired.

        This callback is REQUIRED.
        No errors may occur across this boundary.
        """
        pass

    def on_spi_transfer(self, data_in: bytes) -> bytes:
        """Called to transfer data across the SPI bus.

        This callback is REQUIRED.

        Args:
            data_in(bytes): The data to send.

        Returns(bytes):
            The bytes to send.

        Raises:
            Exception: Implementations are free to raise any Exception. Exceptions are retained
            across the native boundary.
        """
        pass

    def on_finalize(self) -> None:
        """Called when the TCTI is finalized.

        This callback is OPTIONAL.

        Raises:
            Exception: Implementations are free to raise any Exception. Exceptions are retained
            across the native boundary.
        """
        pass

    def on_spi_acquire(self) -> None:
        """Called when the SPI bus needs to be acquired for wait states.

        This callback is REQUIRED for WAIT STATES.
        No errors may occur across this boundary.
        """
        pass

    def on_spi_release(self) -> None:
        """Called when the SPI bus needs to be released for wait states.

        This callback is REQUIRED for WAIT STATES.

        Raises:
            Exception: Implementations are free to raise any Exception. Exceptions are retained
            across the native boundary.
        """
        pass



# SPDX-License-Identifier: BSD-2

from ._libtpm2_pytss import ffi, lib

from .internal.utils import _chkrc
from .constants import TSS2_RC, TPM2_RC
from .TSS2_Exception import TSS2_Exception

import os
from typing import Optional, Tuple, Union


class PollData(object):
    """Initialize a PollData object with OS specific details.

    Initialize a PollData object that holds all OS specific state and metadata information
    for using in the platforms specific asynchronous IO "polling" system. For Linux this
    is Poll, Windows is WaitForSingleObject or other interfaces. Only posix systems currently
    support the events attribute. The system is identified by `os.name`.

    Args:
        fd (int): The File Descriptor(fd) for posix systems or Opque Handle for other systems.
        events (int): The event mask, only for posix systems.

    Returns:
        An instance of the PollData class.
    """

    def __init__(self, fd: int = -1, events: int = -1):
        self._fd = fd
        self._events = events

    @property
    def fd(self) -> int:
        """Gets the File Descriptor or Handle for the asynch I/O wait event.

        Returns:
            The fd or handle.
        """
        return self._fd

    @property
    def handle(self) -> int:
        """Gets the File Descriptor or Handle for the asynch I/O wait event.

        Same as attribute fd.

        Returns:
            The fd or handle.
        """
        return self._fd

    @property
    def events(self) -> int:
        """Gets the Event Mask for the asynch I/O wait event. Suitable for poll.

        Returns:
            The poll event mask.

        Raises:
            NotImplementedError if os.name does not equal "posix".
        """
        if os.name != "posix":
            raise NotImplementedError(
                f"Non POSIX os detected, pollin events not supported, got: {os.name}"
            )
        return self._events


def common_checks(version=1, null_ok=False):
    def decorator(func):
        def wrapper(self, *args, **kwargs):
            def camel_case(s):
                from re import sub

                s = sub(r"(_|-)+", " ", s).title().replace(" ", "")
                return "".join([s[0].lower(), s[1:]])

            if self._v1.version < 1:
                raise TSS2_Exception(TSS2_RC.TCTI_RC_ABI_MISMATCH)

            sub_struct = getattr(self, f"_v{version}")
            if sub_struct is None:
                raise TSS2_Exception(TSS2_RC.TCTI_RC_NOT_IMPLEMENTED)

            method = func.__name__
            method = camel_case(method)
            got_method = getattr(sub_struct, method)
            if not null_ok and got_method == ffi.NULL:
                raise TSS2_Exception(TSS2_RC.TCTI_RC_NOT_IMPLEMENTED)

            try:
                self._clear_exceptions()
                return func(self, *args, **kwargs)
            except Exception as e:
                e = self._get_current_exception(e)
                self._clear_exceptions()
                raise e

        return wrapper

    return decorator


class TCTI:
    """Initialize a TCTI object.

    Initialize a TCTI from a NATIVE instantiated TCTI.

    Args:
        ctx (ffi.CData): A TSS2_TCTI_CONTEXT * variable. This would be returned from a TCTIs
            initialize or TCTILdr routine.

    Returns:
        An instance of a TCTI.
    """

    def __init__(self, ctx: ffi.CData):
        self._v1 = ffi.cast("TSS2_TCTI_CONTEXT_COMMON_V1 *", ctx)
        if self._v1.version == 2:
            self._v2 = ffi.cast("TSS2_TCTI_CONTEXT_COMMON_V2 *", ctx)
        else:
            self._v2 = None
        self._ctx = ctx
        # record the last exception so we can throw across the C boundry without
        # everything becoming an unknown TSS2_Exception(TSS2_RC.TCTI_RC_GENERAL_FAILURE)
        # Normal TCTIs cannot make use of this, by Python TCTIs can. Add it to the base class
        # for subordinate TCTIs to use. This way the TCTI fn calls return the most helpful
        # error.
        self._last_exception = None

    def _set_last_exception(self, exc):
        self._last_exception = exc

    @property
    def _tcti_context(self):
        return self._ctx

    @property
    def magic(self) -> bytes:
        """Returns the MAGIC string of the TCTI.

        Returns:
            The magic byte string.
        """

        # uint64_t in C land by default or let subclass control it
        magic_len = getattr(self, "_magic_len", 8)
        return self._v1.magic.to_bytes(magic_len, "big")

    @property
    def version(self) -> int:
        """Returns the VERSION number of the TCTI.

        This is the TCTI interface version NOT the release
        version of the TCTI. Ie if it implements version 1
        or version 2 of the spec.

        Returns:
            The TCTI version number.
        """

        return self._v1.version

    def _clear_exceptions(self):
        self._last_exception = None

    def _get_current_exception(self, e: Exception):
        x = self._last_exception
        return x if x is not None else e

    @common_checks()
    def transmit(self, command: bytes) -> None:
        """Transmits bytes to the TPM.

        Args:
            command (bytes): The bytes to transmit to the TPM.

        Returns:
            The TCTI version number.

        Raises:
            TSS2_Exception - Underlying TCTI errors
            Exception - Underlying Python TCTIs can return anything.
        """

        cmd = ffi.new("uint8_t []", command)
        clen = len(command)
        _chkrc(self._v1.transmit(self._ctx, clen, cmd))

    @common_checks()
    def receive(self, size: int = 4096, timeout: int = -1) -> bytes:
        """Receives bytes from the TPM.

        Args:
            size (int): The maximum expected response size. Defaults to 4096.
                Negative values infer the default.
            timeout (int): The maximum time to wait for a response in milliseconds.
                Defaults to -1 which will wait indefinitely.

        Returns:
            The TPM response as bytes.

        Raises:
            TSS2_Exception - Underlying TCTI errors
            Exception - Underlying Python TCTIs can return anything.
        """

        if size < 0:
            size = 4096

        resp = ffi.new("uint8_t []", b"\x00" * size)
        rsize = ffi.new("size_t *", size)
        _chkrc(self._v1.receive(self._ctx, rsize, resp, timeout))

        return bytes(ffi.buffer(resp, rsize[0]))

    @common_checks(null_ok=True)
    def finalize(self):
        """Cleans up a TCTI's state and resources."""

        if self._v1.finalize != ffi.NULL:
            self._v1.finalize(self._ctx)
            if self._last_exception:
                e = self._last_exception
                self._clear_exceptions()
                raise e

    @common_checks()
    def cancel(self) -> None:
        """Cancels a current transmit with the TPM.

        Some TCTIs may support the ability to cancel the current
        I/O Operation with the TPM.

        Raises:
            TSS2_Exception - Underlying TCTI errors
            Exception - Underlying Python TCTIs can return anything.
        """

        _chkrc(self._v1.cancel(self._ctx))

    @common_checks()
    def get_poll_handles(self) -> Tuple[PollData]:
        """Gets the poll handles from the TPM.

        Returns:
            A tuple of PollData objects.

        Raises:
            TSS2_Exception - Underlying TCTI errors
            Exception - Underlying Python TCTIs can return anything.
        """

        nhandles = ffi.new("size_t *", 0)
        _chkrc(self._v1.getPollHandles(self._ctx, ffi.NULL, nhandles))
        if nhandles[0] == 0:
            return ()
        handles = ffi.new("TSS2_TCTI_POLL_HANDLE []", nhandles[0])
        _chkrc(self._v1.getPollHandles(self._ctx, handles, nhandles))
        rh = []
        for i in range(0, nhandles[0]):
            if os.name == "posix":
                pd = PollData(handles[i].fd, handles[i].events)
            else:
                pd = PollData(handles[i])
            rh.append(pd)
        return tuple(rh)

    @common_checks()
    def set_locality(self, locality: int) -> None:
        """Sets the locality of the current TCTI connection with the TPM.

        Locality is a value that specifies to the TPM whom is making the
        request. Ie firmware, OS, userspace, etc. For TCTIs and
        TPMs that support this, this interface allows one to set the
        locality.

        Args:
            locality (int): The locality value as an integer.

        Raises:
            TSS2_Exception - Underlying TCTI errors
            Exception - Underlying Python TCTIs can return anything.
        """

        _chkrc(self._v1.setLocality(self._ctx, locality))

    @common_checks(version=2)
    def make_sticky(self, handle: int, sticky: Union[bool, int]) -> None:
        """Makes an object specified by handle not be flushed by a resource manager.

        Resource Managers (RM) MAY flush transient objects when the client disconnects.
        Thus this object would need to be re-established later, eg TPM2_Load command,
        this allows RMs that support the ability to mark this object as non-flushable.

        Raises:
            TSS2_Exception - Underlying TCTI errors
            Exception - Underlying Python TCTIs can return anything.
        """

        hptr = ffi.new("TPM2_HANDLE *", handle)
        _chkrc(self._v2.makeSticky(self._ctx, hptr, sticky))
        return hptr[0]

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.finalize()


# Global callbacks
@ffi.def_extern()
def _tcti_transmit_wrapper(ctx, size, command):
    pi = PyTCTI._cffi_cast(ctx)
    if not hasattr(pi, "do_transmit"):
        return TSS2_RC.TCTI_RC_NOT_IMPLEMENTED
    try:
        pi.do_transmit(bytes(ffi.buffer(command, size)))
    except Exception as e:
        rc = e.rc if isinstance(e, TSS2_Exception) else TSS2_RC.TCTI_RC_GENERAL_FAILURE
        pi._set_last_exception(e)
        return rc

    return TPM2_RC.SUCCESS


@ffi.def_extern()
def _tcti_receive_wrapper(ctx, size, response, timeout):

    # Let the allocator know how much we need.
    pi = PyTCTI._cffi_cast(ctx)
    if response == ffi.NULL:
        size[0] = pi._max_size
        return TPM2_RC.SUCCESS

    if not hasattr(pi, "do_receive"):
        return TSS2_RC.TCTI_RC_NOT_IMPLEMENTED
    try:
        resp = pi.do_receive(timeout)
        max_size = size[0]
        if len(resp) > max_size:
            raise TSS2_Exception(TSS2_RC.TCTI_RC_INSUFFICIENT_BUFFER)

        size[0] = len(resp)
        ffi.memmove(response, resp, len(resp))
    except Exception as e:
        rc = e.rc if isinstance(e, TSS2_Exception) else TSS2_RC.TCTI_RC_GENERAL_FAILURE
        pi._set_last_exception(e)
        return rc

    return TPM2_RC.SUCCESS


@ffi.def_extern()
def _tcti_cancel_wrapper(ctx):
    pi = PyTCTI._cffi_cast(ctx)
    if not hasattr(pi, "do_cancel"):
        return TSS2_RC.TCTI_RC_NOT_IMPLEMENTED
    try:
        pi.do_cancel()
    except Exception as e:
        rc = e.rc if isinstance(e, TSS2_Exception) else TSS2_RC.TCTI_RC_GENERAL_FAILURE
        pi._set_last_exception(e)
        return rc

    return TPM2_RC.SUCCESS


@ffi.def_extern()
def _tcti_get_pollfds_wrapper(ctx, handles, cnt):
    pi = PyTCTI._cffi_cast(ctx)
    if not hasattr(pi, "do_get_poll_handles"):
        return TSS2_RC.TCTI_RC_NOT_IMPLEMENTED
    try:
        # Populate a cache so Python implementors don't have to be called
        # for size and then fd's. FDs should be stable.
        if pi._poll_handle_cache is None:
            pi._poll_handle_cache = pi.do_get_poll_handles()
            # Support callers returning None or list
            if pi._poll_handle_cache is None:
                pi._poll_handle_cache = ()

        # caller wants size for allocation
        if handles == ffi.NULL:
            cnt[0] = len(pi._poll_handle_cache)
        elif cnt[0] < len(pi._poll_handle_cache):
            raise TSS2_RC.TCTI_RC_INSUFFICIENT_BUFFER
        else:
            cnt[0] = len(pi._poll_handle_cache)
            # Enumerate didn't work here
            for i in range(0, cnt[0]):
                pd = pi._poll_handle_cache[i]
                # convert platform agnostic into CData
                if os.name == "posix":
                    handles[i].fd = pd.fd
                    handles[i].events = pd.events
                else:
                    handles[i] = pd.handle
    except Exception as e:
        rc = e.rc if isinstance(e, TSS2_Exception) else TSS2_RC.TCTI_RC_GENERAL_FAILURE
        pi._set_last_exception(e)
        return rc

    return TPM2_RC.SUCCESS


@ffi.def_extern()
def _tcti_set_locality_wrapper(ctx, locality):
    pi = PyTCTI._cffi_cast(ctx)
    if not hasattr(pi, "do_set_locality"):
        return TSS2_RC.TCTI_RC_NOT_IMPLEMENTED
    try:
        pi.do_set_locality(locality)
    except Exception as e:
        rc = e.rc if isinstance(e, TSS2_Exception) else TSS2_RC.TCTI_RC_GENERAL_FAILURE
        pi._set_last_exception(e)
        return rc

    return TPM2_RC.SUCCESS


@ffi.def_extern()
def _tcti_make_sticky_wrapper(ctx, handle, sticky):
    pi = PyTCTI._cffi_cast(ctx)
    if not hasattr(pi, "do_make_sticky"):
        return TSS2_RC.TCTI_RC_NOT_IMPLEMENTED
    try:
        pi.do_make_sticky(handle, bool(sticky))
    except Exception as e:
        rc = e.rc if isinstance(e, TSS2_Exception) else TSS2_RC.TCTI_RC_GENERAL_FAILURE
        pi._set_last_exception(e)
        return rc

    return TPM2_RC.SUCCESS


@ffi.def_extern()
def _tcti_finalize_wrapper(ctx):
    pi = PyTCTI._cffi_cast(ctx)
    if not hasattr(pi, "do_finalize"):
        return

    try:
        pi.do_finalize()
    except Exception as e:
        pi._set_last_exception(e)


class PyTCTI(TCTI):
    """Subclass for implementing a TCTI in Python.

    Extend this object and implement the following methods:
        - def do_transmit(self, command: bytes) -> None
            This method transmits a command buffer to the TPM. This method IS REQUIRED.

        - def do_receive(self, timeout: int) -> bytes:
            This method receives a response from the TPM and returns it. This method IS REQUIRED

        - def do_cancel(self) -> None:
             Cancels an I/O operation with the TPM. This method is OPTIONAL.

        - def do_get_poll_handles(self) -> Optional[Tuple[PollData]]:
             Retrieves PollData objects from the TCTI used for async I/O. This method is OPTIONAL.

        - def do_set_locality(self, locality: int) -> None:
             Sets the locality in which to communicate with the TPM. This method is OPTIONAL.

        - def do_make_sticky(self, handle: int, is_sticky: bool) -> None:
             Makes a handle sticky to persist across client exits with an RM. This method is OPTIONAL.

        - def do_finalize(self) -> None:
             Finalizes a TCTI, this is analogous to close on a file. This method is OPTIONAL.

    Note:
        All methods may throw exceptions as needed.

    Args:
        max_size (int): The size of the response buffer for callers to allocate. Defaults to 4096.
        magic (bytes): The magic value for the TCTI, may aid in debugging. Max length is 8, defaults to b"PYTCTI\x00\x00"

    Returns:
        An instance of the PyTCTI class. It's unusable as is, users should extend it.
    """

    def __init__(self, max_size: int = 4096, magic: bytes = b"PYTCTI\x00\x00"):
        # PYTCTI ASCII FOR MAGIC: echo -n "5059544354490000" | xxd -r -cdata

        if len(magic) > 8:
            raise ValueError(f"Expected magic to be at most 8 bytes, got: {len(magic)}")

        cdata = self._cdata = ffi.new("PYTCTI_CONTEXT *")
        self._max_size = max_size
        self._poll_handle_cache = None
        self._magic_len = len(magic)
        cdata.common.v1.version = 2
        cdata.common.v1.magic = int.from_bytes(magic, "big")
        cdata.common.v1.transmit = lib._tcti_transmit_wrapper
        cdata.common.v1.receive = lib._tcti_receive_wrapper
        cdata.common.v1.cancel = lib._tcti_cancel_wrapper
        cdata.common.v1.getPollHandles = lib._tcti_get_pollfds_wrapper
        cdata.common.v1.setLocality = lib._tcti_set_locality_wrapper
        cdata.common.makeSticky = lib._tcti_make_sticky_wrapper
        cdata.common.v1.finalize = lib._tcti_finalize_wrapper

        # Keep a pointer to this object in the TCTI Context to use later
        # This is how we multiplex N objects through a set of static
        # callbacks. Assign it to an object instance variable to prevent
        # it from getting GC
        cdata.thiz = self._thiz = ffi.new_handle(self)

        opaque = ffi.cast("TSS2_TCTI_CONTEXT *", cdata)
        super().__init__(opaque)

    @staticmethod
    def _cffi_cast(ctx):
        ctx = ffi.cast("PYTCTI_CONTEXT *", ctx)
        return ffi.from_handle(ctx.thiz)

    def do_transmit(self, command: bytes) -> None:
        """This method transmits a command buffer to the TPM. This method IS REQUIRED.

        Args:
            command (bytes): The bytes to send to the TPM.

        Raises:
            NotImplementedError: If a subclass has not implemented this.
            Exception: Implementations are free to raise any Exception. Exceptions are retained
            across the native boundary.
        """

        raise NotImplementedError("Subclass needs to implement do_transmit")

    def do_receive(self, timeout: int) -> bytes:
        """This method receives a response from the TPM and returns it. This method IS REQUIRED.

        Args:
            timeout (int): The timeout in milliseconds to wait for the TPM. Negative values mean
                wait indefinitely.

        Raises:
            NotImplementedError: If a subclass has not implemented this.
            Exception: Implementations are free to raise any Exception. Exceptions are retained
            across the native boundary.
        """

        raise NotImplementedError("Subclass needs to implement do_receive")

    def do_cancel(self) -> None:
        """Cancels an I/O operation with the TPM. This method is OPTIONAL.

        Raises:
            Exception: Implementations are free to raise any Exception. Exceptions are retained
            across the native boundary.
        """
        pass

    def do_get_poll_handles(self) -> Optional[Tuple[PollData]]:
        """Retrieves PollData objects from the TCTI used for async I/O. This method is OPTIONAL.

        Returns:
            The tuple of PollData handles or None.

        Raises:
            Exception: Implementations are free to raise any Exception. Exceptions are retained
            across the native boundary.
        """
        pass

    def do_set_locality(self, locality: int) -> None:
        """Sets the locality in which to communicate with the TPM. This method is OPTIONAL.

        Args:
            locality(int): The locality of communication with the TPM.

        Raises:
            Exception: Implementations are free to raise any Exception. Exceptions are retained
            across the native boundary.
        """
        pass

    def do_make_sticky(self, handle: int, is_sticky: bool) -> None:
        """Makes a handle sticky to persist across client exits with a Resource Manager. This method is OPTIONAL.

        Note: A sticky object is one a RM doesn't flush when the client closes their connection.

        Args:
            handle(int): The TPM handle to make sticky.
            is_sticky(bool): True to make sticky, False to make it not sticky.

        Raises:
            Exception: Implementations are free to raise any Exception. Exceptions are retained
            across the native boundary.
        """
        pass

    def do_finalize(self) -> None:
        """Finalizes a TCTI, this is analogous to close on a file. This method is OPTIONAL.

        Note: Native TCTIs do not return anything and thus cannot raise any errors. Python
        TCTIs MAY raise exceptions across this interface.

        Raises:
            Exception: Implementations are free to raise any Exception. Exceptions are retained
            across the native boundary.
        """
        pass



# SPDX-License-Identifier: BSD-2
from .internal.utils import _lib_version_atleast, _chkrc

if not _lib_version_atleast("tss2-policy", "4.0.0"):
    raise NotImplementedError("tss2-policy not installed or version is less then 4.0.0")

from .types import (
    TPM2B_DIGEST,
    TPMS_NV_PUBLIC,
    TPM2B_NAME,
    TPMT_PUBLIC,
    TPM2B_NONCE,
    TSS2_OBJECT,
    TSS2_POLICY_PCR_SELECTION,
    TPM2_HANDLE,
)
from .constants import TPM2_ALG, ESYS_TR, TSS2_RC, TPM2_RC
from .TSS2_Exception import TSS2_Exception
from ._libtpm2_pytss import ffi, lib
from .ESAPI import ESAPI
from enum import Enum
from typing import Callable, Union


class policy_cb_types(Enum):
    """Policy callback types"""

    CALC_PCR = 0
    CALC_NAME = 1
    CALC_PUBLIC = 2
    CALC_NVPUBLIC = 3
    EXEC_AUTH = 4
    EXEC_POLSEL = 5
    EXEC_SIGN = 6
    EXEC_POLAUTH = 7
    EXEC_POLAUTHNV = 8
    EXEC_POLDUP = 9
    EXEC_POLACTION = 10


@ffi.def_extern()
def _policy_cb_calc_pcr(selection, out_selection, out_digest, userdata):
    """Callback wrapper for policy PCR calculations

    Args:
        selection (struct TSS2_POLICY_PCR_SELECTION): in
        out_selection (struct TPML_PCR_SELECTION): out
        out_digest (struct TPML_DIGEST): out
        userdata (ffi handle): in
    """
    pi = ffi.from_handle(userdata)
    cb = pi._get_callback(policy_cb_types.CALC_PCR)
    if not cb:
        return TSS2_RC.POLICY_RC_NULL_CALLBACK
    try:
        selcopy = ffi.new("TSS2_POLICY_PCR_SELECTION *", selection[0])
        sel = TSS2_POLICY_PCR_SELECTION(_cdata=selcopy)
        cb_selection, cb_digest = cb(sel)
        out_selection.count = cb_selection.count
        for i in range(0, cb_selection.count):
            out_selection.pcrSelections[i] = cb_selection[i]._cdata
        out_digest.count = cb_digest.count
        for i in range(0, cb_digest.count):
            out_digest.digests[i].buffer = cb_digest[i]
            out_digest.digests[i].size = len(cb_digest[i])
    except Exception as e:
        rc = (
            e.rc if isinstance(e, TSS2_Exception) else TSS2_RC.POLICY_RC_GENERAL_FAILURE
        )
        pi._callback_exception = e
        return rc
    return TPM2_RC.SUCCESS


@ffi.def_extern()
def _policy_cb_calc_name(path, name, userdata):
    """Callback wrapper for policy name calculations

    Args:
        path (C string): in
        name (struct TPM2B_DIGEST): out
        userdata (ffi handle): in
    """
    pi = ffi.from_handle(userdata)
    cb = pi._get_callback(policy_cb_types.CALC_NAME)
    if not cb:
        return TSS2_RC.POLICY_RC_NULL_CALLBACK
    try:
        pth = ffi.string(path)
        cb_name = cb(pth)
        name.size = len(cb_name)
        name.name = bytes(cb_name.name)
    except Exception as e:
        rc = (
            e.rc if isinstance(e, TSS2_Exception) else TSS2_RC.POLICY_RC_GENERAL_FAILURE
        )
        pi._callback_exception = e
        return rc
    return TPM2_RC.SUCCESS


@ffi.def_extern()
def _policy_cb_calc_public(path, public, userdata):
    """Callback wrapper for getting the public part for a key path

    Args:
        path (C string): in
        public (struct TPMT_PUBLIC): out
        userdata (ffi handle): in
    """
    pi = ffi.from_handle(userdata)
    cb = pi._get_callback(policy_cb_types.CALC_PUBLIC)
    if not cb:
        return TSS2_RC.POLICY_RC_NULL_CALLBACK
    try:
        pth = ffi.string(path)
        cb_public = cb(pth)
        public.type = cb_public.type
        public.nameAlg = cb_public.nameAlg
        public.objectAttributes = cb_public.objectAttributes
        public.authPolicy.buffer = bytes(cb_public.authPolicy)
        public.authPolicy.size = cb_public.authPolicy.size
        public.parameters = cb_public.parameters._cdata
        public.unique = cb_public.unique._cdata
    except Exception as e:
        rc = (
            e.rc if isinstance(e, TSS2_Exception) else TSS2_RC.POLICY_RC_GENERAL_FAILURE
        )
        pi._callback_exception = e
        return rc
    return TPM2_RC.SUCCESS


@ffi.def_extern()
def _policy_cb_calc_nvpublic(path, nv_index, nv_public, userdata):
    """Callback wrapper for getting the public part for a NV path

    Args:
        path (C string or NULL): in
        nv_index (TPMI_RH_NV_INDEX or zero): in
        nv_public (struct TPMS_NV_PUBLIC): out
        userdata (ffi handle): in
    """
    pi = ffi.from_handle(userdata)
    cb = pi._get_callback(policy_cb_types.CALC_NVPUBLIC)
    if not cb:
        return TSS2_RC.POLICY_RC_NULL_CALLBACK
    try:
        pth = ffi.string(path) if path != ffi.NULL else None
        index = TPM2_HANDLE(nv_index)
        cb_nv_public = cb(pth, index)
        nv_public.nvIndex = cb_nv_public.nvIndex
        nv_public.nameAlg = cb_nv_public.nameAlg
        nv_public.attributes = cb_nv_public.attributes
        nv_public.authPolicy.buffer = bytes(cb_nv_public.authPolicy)
        nv_public.authPolicy.size = cb_nv_public.authPolicy.size
        nv_public.dataSize = cb_nv_public.dataSize
    except Exception as e:
        rc = (
            e.rc if isinstance(e, TSS2_Exception) else TSS2_RC.POLICY_RC_GENERAL_FAILURE
        )
        pi._callback_exception = e
        return rc
    return TPM2_RC.SUCCESS


@ffi.def_extern()
def _policy_cb_exec_auth(name, object_handle, auth_handle, auth_session, userdata):
    """Callback wrapper for getting authorization sessions for a name

    Args:
        name (struct TPM2B_NAME): in
        object_handle (ESYS_TR): out
        auth_handle (ESYS_TR): out
        auth_session (ESYS_TR): out
        userdata (ffi handle): in
    """
    pi = ffi.from_handle(userdata)
    cb = pi._get_callback(policy_cb_types.EXEC_AUTH)
    if not cb:
        return TSS2_RC.POLICY_RC_NULL_CALLBACK
    try:
        nb = ffi.unpack(name.name, name.size)
        name2b = TPM2B_NAME(nb)
        cb_object_handle, cb_auth_handle, cb_auth_session = cb(name2b)
        object_handle[0] = cb_object_handle
        auth_handle[0] = cb_auth_handle
        auth_session[0] = cb_auth_session
    except Exception as e:
        rc = (
            e.rc if isinstance(e, TSS2_Exception) else TSS2_RC.POLICY_RC_GENERAL_FAILURE
        )
        pi._callback_exception = e
        return rc
    return TPM2_RC.SUCCESS


@ffi.def_extern()
def _policy_cb_exec_polsel(
    auth_object, branch_names, branch_count, branch_idx, userdata
):
    """Callback wrapper selection of a policy branch

    Args:
        auth_object (struct TSS2_OBJECT): in
        branch_names (array of C strings): in
        branch_count (int): in
        branch_idx (int): out
        userdata (ffi handle): in
    """
    pi = ffi.from_handle(userdata)
    cb = pi._get_callback(policy_cb_types.EXEC_POLSEL)
    if not cb:
        return TSS2_RC.POLICY_RC_NULL_CALLBACK
    try:
        obj = None
        if auth_object:
            obj = TSS2_OBJECT(handle=auth_object.handle)
        branches = list()
        for i in range(0, branch_count):
            branch = ffi.string(branch_names[i])
            branches.append(branch)
        indx = cb(obj, branches)
        branch_idx[0] = indx
    except Exception as e:
        rc = (
            e.rc if isinstance(e, TSS2_Exception) else TSS2_RC.POLICY_RC_GENERAL_FAILURE
        )
        pi._callback_exception = e
        return rc
    return TPM2_RC.SUCCESS


@ffi.def_extern()
def _policy_cb_exec_sign(
    key_pem,
    public_key_hint,
    key_pem_hash_alg,
    buf,
    buf_size,
    signature,
    signature_size,
    userdata,
):
    """Callback wrapper to signing an operation

    Args:
        key_pem (C string): in
        public_key_hint (C string): in
        key_pem_hash_alg (TPMI_ALG_HASH): in
        buf: (uint8_t array): in
        buf_size (size_t): in
        signature (pointer to uint8_t array): out
        signature_size (pointer to size_t): out
        userdata (ffi handle): in
    """
    pi = ffi.from_handle(userdata)
    cb = pi._get_callback(policy_cb_types.EXEC_SIGN)
    if not cb:
        return TSS2_RC.POLICY_RC_NULL_CALLBACK
    try:
        pem = ffi.string(key_pem)
        key_hint = ffi.string(public_key_hint)
        hash_alg = TPM2_ALG(key_pem_hash_alg)
        b = bytes(ffi.unpack(buf, buf_size))
        cb_signature = cb(pem, key_hint, hash_alg, b)
        signature[0] = ffi.new("char[]", cb_signature)
        signature_size[0] = len(cb_signature)
    except Exception as e:
        rc = (
            e.rc if isinstance(e, TSS2_Exception) else TSS2_RC.POLICY_RC_GENERAL_FAILURE
        )
        pi._callback_exception = e
        return rc
    return TPM2_RC.SUCCESS


@ffi.def_extern()
def _policy_cb_exec_polauth(
    key_public, hash_alg, digest, policy_ref, signature, userdata
):
    """Callback for signing a policy

    Args:
        key_public (struct TPMT_PUBLIC): in
        hash_alg (TPM2_ALG_ID): in
        digest (struct TPM2B_DIGEST): in
        policy_ref (struct TPM2B_NONCE): in
        signature (struct TPMT_SIGNATURE): out
        userdata (ffi handle): in
    """
    pi = ffi.from_handle(userdata)
    cb = pi._get_callback(policy_cb_types.EXEC_POLAUTH)
    if not cb:
        return TSS2_RC.POLICY_RC_NULL_CALLBACK
    try:
        key_pub = TPMT_PUBLIC(_cdata=key_public)
        halg = TPM2_ALG(hash_alg)
        db = ffi.unpack(digest.buffer, digest.size)
        pb = ffi.unpack(policy_ref.buffer, policy_ref.size)
        dig = TPM2B_DIGEST(db)
        polref = TPM2B_NONCE(pb)
        cb_signature = cb(key_pub, halg, dig, polref)
        signature.sigAlg = cb_signature.sigAlg
        signature.signature = cb_signature.signature._cdata
    except Exception as e:
        rc = (
            e.rc if isinstance(e, TSS2_Exception) else TSS2_RC.POLICY_RC_GENERAL_FAILURE
        )
        pi._callback_exception = e
        return rc
    return TPM2_RC.SUCCESS


@ffi.def_extern()
def _policy_cb_exec_polauthnv(nv_public, hash_alg, userdata):
    """Callback wrapper for NV policy authorization

    Args:
        nv_public (struct TPMS_NV_PUBLIC): in
        hash_alg (TPM2_ALG_ID): in
        userdata (ffi handle): in
    """
    pi = ffi.from_handle(userdata)
    cb = pi._get_callback(policy_cb_types.EXEC_POLAUTHNV)
    if not cb:
        return TSS2_RC.POLICY_RC_NULL_CALLBACK
    try:
        nvp = TPMS_NV_PUBLIC(nv_public)
        halg = TPM2_ALG(hash_alg)
        cb(nvp, halg)
    except Exception as e:
        rc = (
            e.rc if isinstance(e, TSS2_Exception) else TSS2_RC.POLICY_RC_GENERAL_FAILURE
        )
        pi._callback_exception = e
        return rc
    return TPM2_RC.SUCCESS


@ffi.def_extern()
def _policy_cb_exec_poldup(name, userdata):
    """Callback wrapper to get name for duplication selection

    Args:
        name (struct TPM2B_NAME): out
        userdata (ffi handle): in
    """
    pi = ffi.from_handle(userdata)
    cb = pi._get_callback(policy_cb_types.EXEC_POLDUP)
    if not cb:
        return TSS2_RC.POLICY_RC_NULL_CALLBACK
    try:
        cb_name = cb()
        name.size = len(cb_name)
        name.name = bytes(cb_name)
    except Exception as e:
        rc = (
            e.rc if isinstance(e, TSS2_Exception) else TSS2_RC.POLICY_RC_GENERAL_FAILURE
        )
        pi._callback_exception = e
        return rc
    return TPM2_RC.SUCCESS


@ffi.def_extern()
def _policy_cb_exec_polaction(action, userdata):
    """Callback wrapper for policy action

    Args:
        action (C string): in
        userdata (ffi handle): in
    """
    pi = ffi.from_handle(userdata)
    cb = pi._get_callback(policy_cb_types.EXEC_POLACTION)
    if not cb:
        return TSS2_RC.POLICY_RC_NULL_CALLBACK
    try:
        ab = ffi.string(action)
        cb(ab)
    except Exception as e:
        rc = (
            e.rc if isinstance(e, TSS2_Exception) else TSS2_RC.POLICY_RC_GENERAL_FAILURE
        )
        pi._callback_exception = e
        return rc
    return TPM2_RC.SUCCESS


class policy(object):
    """Initialize policy object.

    Args:
        policy (Union(bytes, str]): The JSON policy to calculate or execute.
        hash_alg (TPM2_ALG): The hash algorithm to use for policy calculations.

    Returns:
        An instance of the policy object.

    This class implements the policy part of the TCG TSS 2.0 JSON Data Types
    and Policy Language Specification.

    The specification can be found at https://trustedcomputinggroup.org/resource/tcg-tss-json/
    """

    def __init__(self, policy: Union[bytes, str], hash_alg: TPM2_ALG):
        if isinstance(policy, str):
            policy = policy.encode()
        self._policy = policy
        self._hash_alg = hash_alg
        self._callbacks = dict()
        self._callback_exception = None
        self._ctx_pp = ffi.new("TSS2_POLICY_CTX **")
        _chkrc(lib.Tss2_PolicyInit(policy, hash_alg, self._ctx_pp))
        self._ctx = self._ctx_pp[0]
        self._handle = ffi.new_handle(self)
        self._calc_callbacks = ffi.new("TSS2_POLICY_CALC_CALLBACKS *")
        self._exec_callbacks = ffi.new("TSS2_POLICY_EXEC_CALLBACKS *")

    def __enter__(self):
        return self

    def __exit__(self, _type, value, traceback):
        self.close()

    def close(self):
        """Finalize the policy instance"""
        lib.Tss2_PolicyFinalize(self._ctx_pp)
        self._ctx_pp = ffi.NULL
        self._ctx = ffi.NULL

    @property
    def policy(self) -> bytes:
        """bytes: The JSON policy."""
        return self._policy

    @property
    def hash_alg(self) -> TPM2_ALG:
        """TPM2_ALG: The hash algorithm to be used during policy calculcation."""
        return self._hash_alg

    def _get_callback(self, callback_type: policy_cb_types) -> Callable:
        return self._callbacks.get(callback_type)

    def set_callback(
        self, callback_type: policy_cb_types, callback: Union[None, Callable]
    ):
        """Set callback for policy calculaction or execution

        Args:
            callback_type (policy_cb_types): Which callback to set or unset.
            callback (Union[None, Callable]): The callback function to call, or None to remove the callback.

        Raises:
            ValueError
        """
        userdata = self._handle
        if callback is None:
            userdata = ffi.NULL
        update_calc = False
        update_exec = False
        if callback_type == policy_cb_types.CALC_PCR:
            self._callbacks[callback_type] = callback
            self._calc_callbacks.cbpcr = lib._policy_cb_calc_pcr
            self._calc_callbacks.cbpcr_userdata = userdata
            update_calc = True
        elif callback_type == policy_cb_types.CALC_NAME:
            self._callbacks[callback_type] = callback
            self._calc_callbacks.cbname = lib._policy_cb_calc_name
            self._calc_callbacks.cbname_userdata = userdata
            update_calc = True
        elif callback_type == policy_cb_types.CALC_PUBLIC:
            self._callbacks[callback_type] = callback
            self._calc_callbacks.cbpublic = lib._policy_cb_calc_public
            self._calc_callbacks.cbpublic_userdata = userdata
            update_calc = True
        elif callback_type == policy_cb_types.CALC_NVPUBLIC:
            self._callbacks[callback_type] = callback
            self._calc_callbacks.cbnvpublic = lib._policy_cb_calc_nvpublic
            self._calc_callbacks.cbnvpublic_userdata = userdata
            update_calc = True
        elif callback_type == policy_cb_types.EXEC_AUTH:
            self._callbacks[callback_type] = callback
            self._exec_callbacks.cbauth = lib._policy_cb_exec_auth
            self._exec_callbacks.cbauth_userdata = userdata
            update_exec = True
        elif callback_type == policy_cb_types.EXEC_POLSEL:
            self._callbacks[callback_type] = callback
            self._exec_callbacks.cbpolsel = lib._policy_cb_exec_polsel
            self._exec_callbacks.cbpolsel_userdata = userdata
            update_exec = True
        elif callback_type == policy_cb_types.EXEC_SIGN:
            self._callbacks[callback_type] = callback
            self._exec_callbacks.cbsign = lib._policy_cb_exec_sign
            self._exec_callbacks.cbsign_userdata = userdata
            update_exec = True
        elif callback_type == policy_cb_types.EXEC_POLAUTH:
            self._callbacks[callback_type] = callback
            self._exec_callbacks.cbauthpol = lib._policy_cb_exec_polauth
            self._exec_callbacks.cbauthpol_userdata = userdata
            update_exec = True
        elif callback_type == policy_cb_types.EXEC_POLAUTHNV:
            self._callbacks[callback_type] = callback
            self._exec_callbacks.cbauthnv = lib._policy_cb_exec_polauthnv
            self._exec_callbacks.cbauthnv_userdata = userdata
            update_exec = True
        elif callback_type == policy_cb_types.EXEC_POLDUP:
            self._callbacks[callback_type] = callback
            self._exec_callbacks.cbdup = lib._policy_cb_exec_poldup
            self._exec_callbacks.cbdup_userdata = userdata
            update_exec = True
        elif callback_type == policy_cb_types.EXEC_POLACTION:
            self._callbacks[callback_type] = callback
            self._exec_callbacks.cbaction = lib._policy_cb_exec_polaction
            self._exec_callbacks.cbaction_userdata = userdata
            update_exec = True
        else:
            raise ValueError(f"unsupported callback type {callback_type}")

        if update_calc:
            _chkrc(lib.Tss2_PolicySetCalcCallbacks(self._ctx, self._calc_callbacks))
        elif update_exec:
            _chkrc(lib.Tss2_PolicySetExecCallbacks(self._ctx, self._exec_callbacks))

    def execute(self, esys_ctx: ESAPI, session: ESYS_TR):
        """Executes the policy

        Args:
            esys_ctx (ESAPI): The ESAPI instance to use during policy execution.
            session (ESYS_TR): The policy session to use during execution.

        Raises:
            TSS2_Exception or any possible exception from a callback function.
        """
        try:
            _chkrc(lib.Tss2_PolicyExecute(self._ctx, esys_ctx._ctx, session))
        except Exception as e:
            if self._callback_exception is not None:
                raise self._callback_exception
            raise e
        finally:
            self._callback_exception = None

    def calculate(self):
        """Calculate the policy

        Raises:
            TSS2_Exception
        """
        try:
            _chkrc(lib.Tss2_PolicyCalculate(self._ctx))
        except Exception as e:
            if self._callback_exception is not None:
                raise self._callback_exception
            raise e
        finally:
            self._callback_exception = None

    def get_calculated_json(self) -> bytes:
        """Get the calculated policy as JSON

        Returns:
            The calculated JSON policy as bytes

        Raises:
            TSS2_Exception
        """
        size = ffi.new("size_t *")
        _chkrc(lib.Tss2_PolicyGetCalculatedJSON(self._ctx, ffi.NULL, size))

        cjson = ffi.new("uint8_t[]", size[0])
        _chkrc(lib.Tss2_PolicyGetCalculatedJSON(self._ctx, cjson, size))

        return ffi.string(cjson, size[0])

    @property
    def description(self) -> bytes:
        """bytes: The policy description."""
        size = ffi.new("size_t *")
        _chkrc(lib.Tss2_PolicyGetDescription(self._ctx, ffi.NULL, size))

        desc = ffi.new("uint8_t[]", size[0])
        _chkrc(lib.Tss2_PolicyGetDescription(self._ctx, desc, size))
        return ffi.string(desc, size[0])

    def get_calculated_digest(self) -> TPM2B_DIGEST:
        """Get the digest of the calculated policy

        Returns:
            The digest as a TPM2B_DIGEST.

        Raises:
            TSS2_Exception
        """
        dig = ffi.new("TPM2B_DIGEST *")
        _chkrc(lib.Tss2_PolicyGetCalculatedDigest(self._ctx, dig))
        return TPM2B_DIGEST(_cdata=dig[0])



import _cffi_backend

# check that we can load the C bindings,
# if we can't, provide a better message.
try:
    from ._libtpm2_pytss import lib
except ImportError as e:
    parts = e.msg.split(": ", 2)
    if len(parts) != 3:
        raise e
    path, error, symbol = parts
    if error != "undefined symbol":
        raise e
    raise ImportError(
        f"failed to load tpm2-tss bindigs in {path} due to missing symbol {symbol}, "
        + "ensure that you are using the same libraries the python module was built against."
    )

from .ESAPI import ESAPI

try:
    from .FAPI import *
except NotImplementedError:
    # Built on a system lacking FAPI, ignore
    pass
try:
    from .policy import *
except NotImplementedError:
    # Built on a system lacking libpolicy, ignore
    pass
from .TCTILdr import *
from .TCTI import TCTI, PyTCTI, PollData
from .types import *
from .constants import *
from .TSS2_Exception import TSS2_Exception




# SPDX-License-Identifier: BSD-2
# this file is autogenerated during the build

_versions = {
    "tss2-esys": "4.1.3",
    "tss2-fapi": "4.1.3",
    "tss2-policy": "4.1.3",
    "tss2-tcti-spi-helper": "4.1.3",

}



# SPDX-License-Identifier: BSD-2

from .ESAPI import ESAPI
from .constants import ESYS_TR, TPM2_ALG, TPMA_OBJECT, TPM2_ST, TPM2_RH
from .types import (
    TPMT_RSA_DECRYPT,
    TPM2B_DATA,
    TPMT_SIG_SCHEME,
    TPMT_TK_HASHCHECK,
    TPM2B_ECC_POINT,
    TPMT_ASYM_SCHEME,
    TPMT_ECC_SCHEME,
    TPMU_SIG_SCHEME,
)
from .internal.crypto import (
    public_to_key,
    _get_curve,
    _rsa_decrypt_padding_to_scheme,
    _rsa_sign_padding_to_scheme,
    _int_to_buffer,
    _ecc_sign_algorithm_to_scheme,
    _get_digest,
)
from typing import Union
from cryptography.hazmat.primitives.asymmetric import rsa, ec, padding
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric.utils import Prehashed
from cryptography.hazmat.primitives.serialization import (
    Encoding,
    PrivateFormat,
    KeySerializationEncryption,
)


def _compare_schemes(
    in_scheme: Union[TPMT_RSA_DECRYPT, TPMT_SIG_SCHEME], key_scheme: TPMT_SIG_SCHEME
) -> None:
    """Compare a keys scheme and any scheme passed to sign/decrypt functions.

    Raises:
        ValueError: On any scheme mismatch.
    """
    if key_scheme.scheme == TPM2_ALG.NULL:
        return
    if in_scheme.scheme != key_scheme.scheme:
        raise ValueError(
            f"invalid scheme, scheme has {in_scheme.scheme} but key requires {key_scheme.scheme}"
        )
    if in_scheme.scheme == TPM2_ALG.RSAES:
        return
    if isinstance(in_scheme.details, TPMU_SIG_SCHEME):
        halg = in_scheme.details.any.hashAlg
    else:
        halg = in_scheme.details.anySig.hashAlg
    if halg != key_scheme.details.anySig.hashAlg:
        raise ValueError(
            f"digest algorithm mismatch, scheme has {halg} but key requires {key_scheme.details.anySig.hashAlg}"
        )


class tpm_rsa_private_key(rsa.RSAPrivateKey):
    """Interface to a TPM RSA key for use with the cryptography module.

    Args:
        ectx (ESAPI): The ESAPI instance to use.
        handle (ESYS_TR): The key handle.
        session (ESYS_TR): The session to authorize usage of the key, default is ESYS_TR.PASSWORD

    Notes:
        It is recommended to use the :func:`get_digest_algorithm`, :func:`get_decryption_padding` and :func:`get_signature_padding` methods for highest compatibility.

    Raises:
        ValueError: If the key has the restricted bit set or if the handle doesn't reference an RSA key.
    """

    def __init__(
        self, ectx: ESAPI, handle: ESYS_TR, session: ESYS_TR = ESYS_TR.PASSWORD
    ):
        self._handle = handle
        self._session = session
        self._ectx = ectx
        public, _, _ = ectx.read_public(handle)
        self._public = public.publicArea
        if self._public.type != TPM2_ALG.RSA:
            raise ValueError(
                f"invalid key type, expected {TPM2_ALG.RSA}, got {self._public.type}"
            )
        if self._public.objectAttributes & TPMA_OBJECT.RESTRICTED:
            raise ValueError(
                "TPM key does not allow generic signing and/or decryption (object attribute restricted is set)"
            )

    def decrypt(self, ciphertext: bytes, padding: padding) -> bytes:
        """Implements the decrypt interface.

        See :py:meth:`~cryptography.hazmat.primitives.asymmetric.rsa.RSAPrivateKey.decrypt` for documentation.

        Notes:
            If a non-empty label is used with OAEP padding, this will fail.

        Raises:
            ValueError: if the requested padding isn't supported by the key.
        """
        if not self._public.objectAttributes & TPMA_OBJECT.DECRYPT:
            raise ValueError(
                "TPM key does not allow decryption (object attribute decrypt is not set)"
            )
        scheme = TPMT_RSA_DECRYPT()
        _rsa_decrypt_padding_to_scheme(padding, scheme)
        _compare_schemes(scheme, self._public.parameters.rsaDetail.scheme)
        data2b = self._ectx.rsa_decrypt(
            self._handle, ciphertext, scheme, TPM2B_DATA(), session1=self._session
        )
        return bytes(data2b)

    def public_key(self) -> rsa.RSAPublicKey:
        """Get the public key.

        Returns: the public part of the RSA key as a :py:class:`~cryptography.hazmat.primitives.asymmetric.rsa.RSAPublicKey`.
        """
        return public_to_key(self._public)

    @property
    def key_size(self) -> int:
        """The RSA key size"""
        return self._public.parameters.rsaDetail.keyBits

    def get_digest_algorithm(self) -> hashes.HashAlgorithm:
        """Get an usable digest algorithm for use with the key.

        If any scheme with a specified digest algorithm is specified return that algorithm.
        Otherwise the name digest algorithm is returned.

        The returned digest algorithm can be used with different cryptography functions.

        Returns:
            The digest algorithm as a :py:class:`~cryptography.hazmat.primitives.hashes.HashAlgorithm` subclass.

        Raises:
            ValueError: If the digest algorithm is not supported.
        """
        if self._public.parameters.rsaDetail.scheme.scheme in (
            TPM2_ALG.RSASSA,
            TPM2_ALG.RSAPSS,
            TPM2_ALG.OAEP,
        ):
            tpm_alg = self._public.parameters.rsaDetail.scheme.details.anySig.hashAlg
        else:
            tpm_alg = self._public.nameAlg
        halg = _get_digest(tpm_alg)
        if halg is None:
            raise ValueError(f"unsupported digest algorithm {tpm_alg}")
        return halg

    def get_decryption_padding(self) -> padding.AsymmetricPadding:
        """Get a padding configuration for use with the decrypt method.

        If the key has a scheme specified, use that scheme.
        Otherwise, use OAEP as the default.

        Returns:
            An instance of :py:class:`~cryptography.hazmat.primitives.asymmetric.padding.AsymmetricPadding`.

        Raises:
            ValueError: If the either the scheme or digest algorithm is unsupported.
        """
        if self._public.parameters.asymDetail.scheme.scheme == TPM2_ALG.NULL:
            scheme = TPMT_ASYM_SCHEME(scheme=TPM2_ALG.OAEP)
            scheme.details.anySig.hashAlg = self._public.nameAlg
        else:
            scheme = self._public.parameters.asymDetail.scheme
        if scheme.scheme == TPM2_ALG.OAEP:
            algorithm = self.get_digest_algorithm()
            decrypt_padding = padding.OAEP(
                mgf=padding.MGF1(algorithm=algorithm()),
                algorithm=algorithm(),
                label=b"",
            )
        elif scheme.scheme == TPM2_ALG.RSAES:
            decrypt_padding = padding.PKCS1v15()
        else:
            raise ValueError(f"unsupported decryption scheme {scheme.scheme}")
        return decrypt_padding

    def get_signature_padding(self) -> padding.AsymmetricPadding:
        """Get a padding configuration for use with the sign method.

        If the key has a scheme specified, use that scheme.
        Otherwise, use PSS as the default.

        Returns:
          An instance of :py:class:`~cryptography.hazmat.primitives.asymmetric.padding.AsymmetricPadding`.

        Raises: ValueError if the either the scheme or digest algorithm is unsupported.
        """
        if self._public.parameters.asymDetail.scheme.scheme == TPM2_ALG.NULL:
            scheme = TPMT_ASYM_SCHEME(scheme=TPM2_ALG.RSAPSS)
            scheme.details.anySig.hashAlg = self._public.nameAlg
        else:
            scheme = self._public.parameters.asymDetail.scheme
        if scheme.scheme == TPM2_ALG.RSAPSS:
            algorithm = self.get_digest_algorithm()
            sign_padding = padding.PSS(
                mgf=padding.MGF1(algorithm=algorithm()),
                salt_length=padding.PSS.DIGEST_LENGTH,
            )
        elif scheme.scheme == TPM2_ALG.RSASSA:
            sign_padding = padding.PKCS1v15()
        else:
            raise ValueError(f"unsupported signature scheme {scheme.scheme}")
        return sign_padding

    def sign(
        self,
        data: bytes,
        padding: padding,
        algorithm: Union[hashes.HashAlgorithm, Prehashed],
    ) -> bytes:
        """Implements the sign interface.

        See :py:meth:`~cryptography.hazmat.primitives.asymmetric.rsa.RSAPrivateKey.sign` for documentationen.

        Notes:
            For PSS padding, the salt length should be set to the length of the digest as that is the only setup the TPM uses.

        Raises:
            ValueError: If the requested padding isn't supported by the key or the sign_encrypt bit isn't set.
        """
        if not self._public.objectAttributes & TPMA_OBJECT.SIGN_ENCRYPT:
            raise ValueError(
                "TPM key does not allow signing (object attribute sign_encrypt is not set)"
            )
        if isinstance(algorithm, Prehashed):
            raise ValueError("Prehashed data is not supported")
        scheme = TPMT_SIG_SCHEME()
        _rsa_sign_padding_to_scheme(padding, type(algorithm), scheme)
        _compare_schemes(scheme, self._public.parameters.rsaDetail.scheme)
        h = hashes.Hash(algorithm)
        h.update(data)
        digest = h.finalize()
        validation = TPMT_TK_HASHCHECK(tag=TPM2_ST.HASHCHECK, hierarchy=TPM2_RH.NULL)
        tpm_sig = self._ectx.sign(
            self._handle, digest, scheme, validation, session1=self._session
        )
        return bytes(tpm_sig)

    def private_numbers(self) -> None:
        """Always raises a NotImplementedError."""
        raise NotImplementedError()

    def private_bytes(
        self,
        encoding: Encoding,
        format: PrivateFormat,
        encryption_algorithm: KeySerializationEncryption,
    ) -> None:
        """Always raises a NotImplementedError."""
        raise NotImplementedError()

    def __copy__(self) -> "tpm_rsa_private_key":
        """Returns a shallow copy of the private key."""
        return tpm_rsa_private_key(
            ectx=self._ectx, handle=self._handle, session=self._session
        )


class tpm_ecc_private_key(ec.EllipticCurvePrivateKey):
    """Interface to a TPM ECC key for use with the cryptography module.

    Args:
        ectx (ESAPI): The ESAPI instance to use.
        handle (ESYS_TR): The key handle.
        session (ESYS_TR): The session to authorize usage of the key, default is ESYS_TR.PASSWORD

    Notes:
        It is recommended to use the :func:`get_digest_algorithm` and :func:`get_signature_algorithm` methods for highest compatibility.

    Raises:
        ValueError: If the key has the restricted bit set, the curve isn't supported or if the handle doesn't reference an ECC key.
    """

    def __init__(
        self, ectx: ESAPI, handle: ESYS_TR, session: ESYS_TR = ESYS_TR.PASSWORD
    ):
        self._handle = handle
        self._session = session
        self._ectx = ectx
        public, _, _ = ectx.read_public(handle)
        self._public = public.publicArea
        if self._public.type != TPM2_ALG.ECC:
            raise ValueError(
                f"invalid key type, expected {TPM2_ALG.ECC}, got {self._public.type}"
            )
        if self._public.objectAttributes & TPMA_OBJECT.RESTRICTED:
            raise ValueError(
                "TPM key does not allow generic signing and/or decryption (object attribute restricted is set)"
            )
        cid = _get_curve(self._public.parameters.eccDetail.curveID)
        if cid is None:
            raise ValueError(
                f"unsupported curve {self._public.parameters.eccDetail.curveID}"
            )
        self._curve = cid

    def exchange(
        self, algorithm: ec.ECDH, peer_public_key: ec.EllipticCurvePublicKey
    ) -> bytes:
        """Implements the exchange interface.

        See :py:meth:`~cryptography.hazmat.primitives.asymmetric.ec.EllipticCurvePrivateKey.exchange` for documentationen.

        Raises:
            ValueError: If the curves does not match or the decrypt bit isn't set.
        """
        if not self._public.objectAttributes & TPMA_OBJECT.DECRYPT:
            raise ValueError(
                "TPM key does not allow ECDH key exchange (object attribute decrypt is not set)"
            )
        if type(peer_public_key.curve) != type(self.curve):
            raise ValueError(
                f"curve mismatch for peer key, got {peer_public_key.curve.name}, expected {self.curve.name}"
            )
        scheme = TPMT_SIG_SCHEME(scheme=TPM2_ALG.ECDH)
        _compare_schemes(scheme, self._public.parameters.eccDetail.scheme)
        in_point = TPM2B_ECC_POINT()
        nums = peer_public_key.public_numbers()
        _int_to_buffer(nums.x, in_point.point.x)
        _int_to_buffer(nums.y, in_point.point.y)

        out_point = self._ectx.ecdh_zgen(self._handle, in_point, session1=self._session)
        return bytes(out_point.point.x)

    def public_key(self) -> ec.EllipticCurvePublicKey:
        """Get the public key.

        Returns: the public part of the ECC key as a :py:class:`~cryptography.hazmat.primitives.asymmetric.ec.EllipticCurvePublicKey`
        """
        return public_to_key(self._public)

    def get_digest_algorithm(self) -> hashes.HashAlgorithm:
        """Get an usable digest algorithm for use with the key.

        If any scheme with a specified digest algorithm is specified return that algorithm.
        Otherwise the name digest algorithm is returned.

        The returned digest algorithm can be used with different cryptography functions.

        Returns:
            The digest algorithm as a :py:class:`~cryptography.hazmat.primitives.hashes.HashAlgorithm` subclass.

        Raises:
            ValueError: If the digest algorithm is not supported.
        """
        if self._public.parameters.eccDetail.scheme.scheme == TPM2_ALG.ECDSA:
            tpm_alg = self._public.parameters.eccDetail.scheme.details.anySig.hashAlg
        else:
            tpm_alg = self._public.nameAlg
        halg = _get_digest(tpm_alg)
        if halg is None:
            raise ValueError(f"unsupported digest algorithm {tpm_alg}")
        return halg

    def get_signature_algorithm(self) -> ec.EllipticCurveSignatureAlgorithm:
        """Get a padding configuration for use with the sign method.

        If the key has a scheme specified, use that scheme.
        Otherwise, use ECDSA as the default

        Returns: an instance of :py:class:`~cryptography.hazmat.primitives.asymmetric.ec.EllipticCurveSignatureAlgorithm`

        Raises:
            ValueError: If the either the scheme or digest algorithm is unsupported.
        """
        if self._public.parameters.eccDetail.scheme.scheme == TPM2_ALG.NULL:
            scheme = TPMT_ECC_SCHEME(scheme=TPM2_ALG.ECDSA)
            scheme.details.anySig.hashAlg = self._public.nameAlg
        else:
            scheme = self._public.parameters.eccDetail.scheme
        if scheme.scheme == TPM2_ALG.ECDSA:
            algorithm = self.get_digest_algorithm()
            sig_alg = ec.ECDSA(algorithm())
        else:
            raise ValueError(f"unsupported signature scheme {scheme.scheme}")
        return sig_alg

    def sign(
        self, data: bytes, signature_algorithm: ec.EllipticCurveSignatureAlgorithm
    ) -> bytes:
        """Implements the sign interface.

        See :py:meth:`~cryptography.hazmat.primitives.asymmetric.ec.EllipticCurvePrivateKey.sign`: for documentation.

        Raises:
            ValueError: if the requested signature algorithm isn't supported by the key or the sign_encrypt bit isn't set.
        """
        if not self._public.objectAttributes & TPMA_OBJECT.SIGN_ENCRYPT:
            raise ValueError(
                "TPM key does not allow signing (object attribute sign_encrypt is not set)"
            )
        algorithm = signature_algorithm.algorithm
        if isinstance(algorithm, Prehashed):
            raise ValueError("Prehashed data is not supported")
        scheme = TPMT_SIG_SCHEME()
        _ecc_sign_algorithm_to_scheme(signature_algorithm, scheme)
        _compare_schemes(scheme, self._public.parameters.eccDetail.scheme)
        h = hashes.Hash(algorithm)
        h.update(data)
        digest = h.finalize()
        validation = TPMT_TK_HASHCHECK(tag=TPM2_ST.HASHCHECK, hierarchy=TPM2_RH.NULL)
        tpm_sig = self._ectx.sign(
            self._handle, digest, scheme, validation, session1=self._session
        )
        return bytes(tpm_sig)

    @property
    def curve(self) -> ec.EllipticCurve:
        """The ECC curve."""
        return self._curve()

    @property
    def key_size(self) -> int:
        """The ECC key size."""
        return self.public_key().key_size

    def private_numbers(self) -> None:
        """Always raises a NotImplementedError."""
        raise NotImplementedError()

    def private_bytes(
        self,
        encoding: Encoding,
        format: PrivateFormat,
        encryption_algorithm: KeySerializationEncryption,
    ) -> None:
        """Always raises a NotImplementedError."""
        raise NotImplementedError()

    def __copy__(self) -> "tpm_ecc_private_key":
        """Returns a shallow copy of the private key."""
        return tpm_ecc_private_key(
            ectx=self._ectx, handle=self._handle, session=self._session
        )
